import numpy as np
from openmdao.api import Group, Problem, Component, IndepVarComp, Newton, ScipyGMRES

num_points = 5
point_dim = 2
num_lengthts = 4


class GeometricalConstraints(Component):
    def __init__(self):
        super().__init__()
        self.deriv_options['type'] = 'cs'

        #self.add_param('d', val=np.array([0.04421, 0.09219]))
        #self.add_param('D', val=0.1364)
        self.add_param('D', val=0.1364)
        self.add_param('len3', val=0.5636)
        self.add_param('len2', val=0.1940)
        self.add_param('len0', val=.52999)
        #self.add_param('len0', val=.25)

        # "FASIT"
        #self.add_param('len1', val=02.622)

        self.add_output('len_list', shape=num_lengthts)

    def solve_nonlinear(self, params, unknowns, resids):
        #d = params['d']
        len1 = params['D']/2 + params['len2']

        unknowns['len_list'] = res = np.array([
                                        params['len0'],
                                        len1,
                                        params['len2'],
                                        params['len3']
                                    ])

        print('lengths: ', res)


class PointsCalculations(Component):
    def __init__(self):
        super().__init__()
        self.deriv_options['type'] = 'cs'

        self.add_param('alpha', val=np.array([2.365, 1.2217, 2.356, 0.7941]))
        self.add_param('x_ref', val=np.array([0.3959, -0.1262]))
        self.add_param('len_list', shape=num_lengthts)

        self.add_output('points', shape=(num_points, point_dim))

    def solve_nonlinear(self, params, unknowns, resids):
        x = params['x_ref']
        p_local = [self._pol2cart(rho, phi) for rho, phi
                   in zip(params['len_list'], params['alpha'])]

        points = [(0, 0), p_local[2], p_local[1], p_local[0], p_local[3]]

        # offset rocker base:
        points[3] = points[3] + points[2]

        unknowns['points'] = np.array([p+x for p in points])

    @staticmethod
    def _pol2cart(rho, phi):
        x = rho * np.cos(phi)
        y = rho * np.sin(phi)
        return np.array((x, y))


class PointMapper(Component):

    def __init__(self):
        super().__init__()

        self.add_param('points', shape=(num_points, point_dim))

        self.add_output('P0', shape=3)
        self.add_output('P3', shape=3)
        self.add_output('P4', shape=3)
        self.add_output('P5', shape=3)
        self.add_output('P6', shape=3)

    def solve_nonlinear(self, params, unknowns, resids):
        z0 = np.array([0.0])
        z1 = np.array([-0.06])

        unknowns['P0'] = np.concatenate((params['points'][1], z0))
        unknowns['P3'] = np.concatenate((params['points'][3], z1))
        unknowns['P4'] = np.concatenate((params['points'][2], z1))
        unknowns['P5'] = np.concatenate((params['points'][0], z0))
        unknowns['P6'] = np.concatenate((params['points'][4], z0))


class DesignSpaceTransformer(Group):
    def __init__(self):
        super().__init__()

        self.add('constraints', GeometricalConstraints())
        self.add('transformations', PointsCalculations(), promotes=['points'])
        self.add('mapped_points', PointMapper(), promotes=['points'])

        self.connect('constraints.len_list', 'transformations.len_list')

        #self.nl_solver = Newton()
        #self.nl_solver.options['maxiter'] = 100
        #self.ln_solver = ScipyGMRES()


if __name__ == '__main__':
    root = DesignSpaceTransformer()
    prob = Problem(root)
    prob.setup()
    prob.print_all_convergence(level=2, depth=1)
    prob.run()

    for n in ['transformations.len0', 'transformations.len1', 'transformations.len2', 'transformations.len3']:
        print(n, ': ', prob[n])

    print('POINTS:', prob['transformations.len0'])
    for i, p in enumerate(prob['points']):
        print('point ', i, ' :', p)
