from openmdao.api import Component, Group, Problem, Newton, ScipyGMRES


class Parabola(Component):
    """Evaluates y = 3x^2 - 5"""

    def __init__(self):
        super(Parabola, self).__init__()

        self.add_param('l_x', 1.0)
        self.add_output('l_y', 0.0)

        # User can change these.
        self.slope = -2.0
        self.intercept = 4.0

        self.add_param('p_x', 1.0)
        self.add_output('p_y', 0.0)

        # User can change these.
        self.a = 3.0
        self.b = 0.0
        self.c = -5.0

        self.deriv_options['type'] = 'cs'

    def solve_nonlinear(self, params, unknowns, resids):
        """ y = 3x^2 - 5 """

        x = params['l_x']
        m = self.slope
        b = self.intercept

        unknowns['l_y'] = m*x + b

        x = params['p_x']
        a = self.a
        b = self.b
        c = self.c

        unknowns['p_y'] = a*x**2 + b*x + c


class Balance(Component):
    """Evaluates the residual y1-y2"""

    def __init__(self):
        super(Balance, self).__init__()

        self.add_param('y1', 0.0)
        self.add_param('y2', 0.0)
        self.add_state('x', 5.0)

        self.deriv_options['type'] = 'cs'

    def solve_nonlinear(self, params, unknowns, resids):
        """This component does no calculation on its own. It mainly holds the
        initial value of the state. An OpenMDAO solver outside of this
        component varies it to drive the residual to zero."""
        pass

    def apply_nonlinear(self, params, unknowns, resids):
        """ Report the residual y1-y2 """

        y1 = params['y1']
        y2 = params['y2']

        resids['x'] = y1 - y2


top = Problem()
root = top.root = Group()
#root.add('line', Line())
root.add('parabola', Parabola())
root.add('bal', Balance())

root.connect('parabola.l_y', 'bal.y1')
root.connect('parabola.p_y', 'bal.y2')
root.connect('bal.x', 'parabola.l_x')
root.connect('bal.x', 'parabola.p_x')

root.nl_solver = Newton()
root.ln_solver = ScipyGMRES()

top.setup()

top.run()
print('Solution x=%.2f, line.y=%.2f, parabola.y=%.2f' % (top['bal.x'], top['parabola.l_y'], top['parabola.p_y']))
