from collections import OrderedDict as od
from openmdao.api import Group, Component
import numpy as np


class AggregatedGroup(Group):

    class Aggregator(Component):

        def __init__(self, aggregatees):
            super().__init__()
            self.param_names = aggregatees
            for name in self.param_names:
                self.add_param(name, val='', pass_by_obj=True)

            self.add_output('dict', val='', pass_by_obj=True)

        def solve_nonlinear(self, params, unknowns, resids):
            unknowns['dict'] = {name: params[name] for name in self.param_names}

    def create_aggregator(self):
        comp_names = [c.name for c in self.subsystems()]
        self.add('aggregator', AggregatedGroup.Aggregator(comp_names), promotes=['dict'])
        for name in comp_names:
            self.connect('{}.dict'.format(name), 'aggregator.{}'.format(name))

    def get_name_of(self, child):
        for name, subsys in self._subsystems.items():
            if subsys == child:
                return name
            elif isinstance(subsys, Group):
                subname = subsys.get_name_of(child)
                if subname:
                    return '{}.{}'.format(name, subname)
        return None

class AggregatedComp(Component):
    def __init__(self, desc):
        super().__init__()
        self.add_output('dict', val='', pass_by_obj=True)
        self.desvar = None

    # noqa
    def add_param(self, name, val=None, *, auto_desvar=False, **kwargs):

        if auto_desvar:
            if type(val) == dict:
                self.desvar = ((name, val['default'], val['min'], val['max']))

                val = val['default']

            super().add_param(name, val=np.array(val), **kwargs)

        else:
            super().add_param(name, val, **kwargs)





class Point(AggregatedComp):
    def __init__(self, desc):
        super().__init__(desc)
        self.id = desc['id']
        self.name = desc['name'].replace('-', '_')

        self.add_param('position', val=desc['pos'], auto_desvar=True)

    def solve_nonlinear(self, params, unknowns, resids):
        unknowns['dict'] = {'name': self.name,
                            'id': self.id,
                            'pos': params['position']}


    @staticmethod
    def parse_points(desc):
        points = AggregatedGroup()

        for point_desc in desc:
            point = Point(point_desc)
            points.add(point.name, point)

        return points


class Member(AggregatedComp):
    def __init__(self, desc, id):
        super().__init__(desc)
        self.name = desc['name'].replace(' ', '_')
        self.id = id

        self.add_param('cross_section', val=desc['cross-section'], pass_by_obj=True)
        self.add_param('dimensions', val=desc['dimensions'], auto_desvar=True)

    def solve_nonlinear(self, params, unknowns, resids):
        unknowns['dict'] = {'name': self.name,
                            'cross_section': params['cross_section'],
                            'dimensions': params['dimensions'],
                            'link_name': self.link_name,
                            'link_id': self.link_id,
                            'member_id': self.id}

    @staticmethod
    def parse_members(desc):
        link = AggregatedGroup()

        for i, member_desc in enumerate(desc):
            member = Member(member_desc, i)
            link.add(member.name, member)

        link.create_aggregator()

        return link

    @staticmethod
    def parse_links(desc):
        links = AggregatedGroup()

        for links_desc in desc:
            link = Member.parse_members(links_desc['members'])
            link.id = links_desc['id']
            link.name = 'link{}'.format(link.id)
            links.add(link.name, link)

            for c in link.subsystems():
                c.link_name = link.name
                c.link_id = link.id

        return links


class Constraint(AggregatedComp):
    def __init__(self, desc):
        super().__init__(desc)
        self.id = desc['id']

        constraint_type = desc['type']
        self.add_param('type', val=constraint_type, pass_by_obj=True)

        fixed_dofs = desc['fixed-dofs'] if constraint_type == 'free' else None

        self.add_param('fixed_dofs', val=fixed_dofs, pass_by_obj=True)

        self.add_param('joint_direction', val=desc['joint-direction'], auto_desvar=True)

        self.point = desc['point']['id']
        self.add_param('point', val='', pass_by_obj=True)

        self.linkparam_names = od()
        for i, link_desc in enumerate(desc['links']):
            if link_desc:
                param_name = 'link{}'.format(i)
                link_id = link_desc['id']
                self.linkparam_names[link_id] = param_name
                self.add_param(param_name, val='', pass_by_obj=True)

    def solve_nonlinear(self, params, unknowns, resids):
        unknowns['dict'] = {'name': self.name,
                            'point': params['point']['id'],
                            'type': params['type'],
                            'links': list(self.linkparam_names.keys()),
                            'joint-dir': params['joint_direction'],
                            'fixed-dofs': params['fixed_dofs']
                            }

    @staticmethod
    def parse_constraints(desc):
        constraints = AggregatedGroup()

        for i, constraint_desc in enumerate(desc):
            constraint = Constraint(constraint_desc)
            constraint.name = 'constraint{}'.format(i)
            constraints.add(constraint.name, constraint)

        return constraints


class Load(AggregatedComp):
    def __init__(self, desc):
        super().__init__(desc)
        self.id = desc['id']
        self.type = desc['type']

        self.point = desc['point']['id']
        self.add_param('point', val='', pass_by_obj=True)

        self.link = desc['loaded-link']['id']
        self.add_param('link', val='', pass_by_obj=True)

        self.add_param('magnitude', val=desc['magnitude'], pass_by_obj=True)
        self.add_param('direction', val=desc['direction'], auto_desvar=True)


    def solve_nonlinear(self, params, unknowns, resids):
        unknowns['dict'] = {'point': params['point']['id'],
                            'type': self.type,
                            'link': self.link,
                            'direction': params['direction'],
                            'magnitude': params['magnitude']
                            }

    @staticmethod
    def parse_loads(desc):
        loads = AggregatedGroup()

        for i, load_desc in enumerate(desc):
            load = Load(load_desc)
            load.name = 'load{}'.format(i)
            loads.add(load.name, load)

        return loads


class NodeCoupling(AggregatedComp):
    def __init__(self, desc):
        super().__init__(desc)
        self.id = desc['id']
        self.type = desc['type']

        self.points = [p['id'] for p in desc['points']]
        self.add_param('from_point', val='', pass_by_obj=True)
        self.add_param('to_point', val='', pass_by_obj=True)

        self.add_param('constant', val=desc['constant'])

        self.linkparam_names = od()
        for i, link_desc in enumerate(desc['links']):
            if link_desc:
                param_name = 'link{}'.format(i)
                link_id = link_desc['id']
                self.linkparam_names[link_id] = param_name
                self.add_param(param_name, val='', pass_by_obj=True)
            else:
                self.linkparam_names[None] = None

    def solve_nonlinear(self, params, unknowns, resids):
        unknowns['dict'] = {'points': self.points,
                            'type': self.type,
                            'links': list(self.linkparam_names.keys()),
                            'constant': params['constant']
                            }

    @staticmethod
    def parse_nodecouplings(desc):
        couplings = AggregatedGroup()

        for i, coupling_desc in enumerate(desc):
            coupling = NodeCoupling(coupling_desc)
            coupling.name = 'NodeCoupling{}'.format(i)
            couplings.add(coupling.name, coupling)

        return couplings
