import yaml
from openmdao.api import Problem, view_model, view_connections

from parser_classes import AggregatedGroup, Member, Point, Constraint, Load, NodeCoupling
from file_writer import write_files
from desvar_creator import find_desvars


def _enumerate(desc):
    for array in desc.values():
        for i, elem in enumerate(array):
            elem['id'] = i


def view_group(g):
    prob = Problem(root=g)
    prob.setup()
    prob.run()
    view_connections(g)
    view_model(prob)
    return prob


def make_connections(group):

    point_names = {p.id: p.name for p in group.points.components()}
    link_names = {l.id: l.name for l in group.links.subgroups()}

    constraints = list(group.constraints.components())
    loads = list(group.loads.components())
    couplings = list(group.couplings.components())

    create_aggregator(group)

    connect_constraints(group, constraints, link_names, point_names)
    connect_loads(group, loads, link_names, point_names)
    connect_couplings(group, couplings, link_names, point_names)


def create_aggregator(group):
    group.links.create_aggregator()
    group.points.create_aggregator()
    group.constraints.create_aggregator()
    group.loads.create_aggregator()
    group.couplings.create_aggregator()
    group.create_aggregator()


def connect_couplings(group, couplings, link_names, point_names):
    for coupling in couplings:
        links = ((link_id, linkparam_name) for link_id, linkparam_name
                 in coupling.linkparam_names.items() if linkparam_name)

        for link_id, linkparam_name in links:
            group.connect('links.{}.dict'.format(link_names[link_id]),
                          'couplings.{}.{}'.format(coupling.name, linkparam_name))

        group.connect('points.{}.dict'.format(point_names[coupling.points[0]]),
                      'couplings.{}.from_point'.format(coupling.name))
        group.connect('points.{}.dict'.format(point_names[coupling.points[1]]),
                      'couplings.{}.to_point'.format(coupling.name))


def connect_loads(group, loads, link_names, point_names):
    for load in loads:
        group.connect('points.{}.dict'.format(point_names[load.point]),
                      'loads.{}.point'.format(load.name))
        group.connect('links.{}.dict'.format(link_names[load.link]),
                      'loads.{}.link'.format(load.name))


def connect_constraints(group, constraints, link_names, point_names):
    for cons in constraints:
        group.connect('points.{}.dict'.format(point_names[cons.point]),
                      'constraints.{}.point'.format(cons.name))

        for link_id, linkparam_name in cons.linkparam_names.items():
            group.connect('links.{}.dict'.format(link_names[link_id]),
                          'constraints.{}.{}'.format(cons.name, linkparam_name))


def parsefile(f):
    with open(f) as fp:
        mech_desc = yaml.load(fp)
        _enumerate(mech_desc)

        g = AggregatedGroup()
        g.add('points', Point.parse_points(mech_desc['points']))
        g.add('links', Member.parse_links(mech_desc['links']))
        g.add('constraints', Constraint.parse_constraints(mech_desc['constraints']))
        g.add('loads', Load.parse_loads(mech_desc['loads']))
        g.add('couplings', NodeCoupling.parse_nodecouplings(mech_desc['node-coupling']))

        make_connections(g)
        d = find_desvars(g)

        return g, d


if __name__ == '__main__':
    from random import uniform

    g, d = parsefile('four-bar.yaml')
    prob = Problem(root=g)
    prob.setup()

    # name, a, b = d[1]
    # print('setting ', name)
    # prob[name][1] = uniform(a[1], b[1])
    # prob[name][0] = uniform(a[0], b[0])

    prob.run()
    view_connections(g)
    view_model(prob)

    write_files(prob['dict'])
